'''
    DeNova's open source libraries.

    All files in this directory tree are open source,
    licensed under GPLv3 <http://www.gnu.org/licenses/>.

    Copyright 2009-2020 DeNova
    Last modified: 2020-09-22
'''
