'''
    Extra functions/templates, etc. when you have a django app.
    django_addons is named to avoid conflict with the django pacakge.

    Copyright 2009-2020 DeNova
    Last modified: 2020-10-02

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''
